from typing import List, Optional
from deply.models.dependency import Dependency
from deply.models.violation import Violation
from deply.models.violation_types import ViolationType
from deply.rules.base_rule import BaseRule


class DependencyRule(BaseRule):
    VIOLATION_TYPE = ViolationType.DISALLOWED_DEPENDENCY

    def __init__(self, layer_name: str, disallowed_layers: List[str]):
        self.layer_name = layer_name
        self.disallowed_layers = set(disallowed_layers)

    def check(self, source_layer: str, target_layer: str, dependency: Dependency) -> Optional[Violation]:
        if source_layer != self.layer_name:
            return None
        if target_layer not in self.disallowed_layers:
            return None
        message = (
            f"Layer '{source_layer}' is not allowed to depend on layer '{target_layer}'. "
            f"Dependency type: {dependency.dependency_type}."
        )
        return Violation(
            file=dependency.code_element.file,
            element_name=dependency.code_element.name,
            element_type=dependency.code_element.element_type,
            line=dependency.line,
            column=dependency.column,
            message=message,
            violation_type=self.VIOLATION_TYPE,
            dependency=dependency,
        )



import ast
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

import yaml

from deply.collectors import FileRegexCollector, ClassInheritsCollector
from deply.collectors.bool_collector import BoolCollector
from deply.collectors.class_name_regex_collector import ClassNameRegexCollector
from deply.collectors.decorator_usage_collector import DecoratorUsageCollector
from deply.collectors.directory_collector import DirectoryCollector
from deply.main import main


class TestCollectors(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.test_project_dir = Path(self.test_dir) / 'test_project'
        self.test_project_dir.mkdir()

        # Create directories
        (self.test_project_dir / 'controllers').mkdir()
        (self.test_project_dir / 'models').mkdir()
        (self.test_project_dir / 'services').mkdir()
        (self.test_project_dir / 'excluded_folder_name').mkdir()
        (self.test_project_dir / 'utilities').mkdir()

        # Create files in controllers
        base_controller_py = self.test_project_dir / 'controllers' / 'base_controller.py'
        base_controller_py.write_text('class BaseController:\n    pass\n')

        user_controller_py = self.test_project_dir / 'controllers' / 'user_controller.py'
        user_controller_py.write_text('@login_required\nclass UserController(BaseController):\n    pass\n')

        # Create files in models
        base_model_py = self.test_project_dir / 'models' / 'base_model.py'
        base_model_py.write_text('class BaseModel:\n    pass\n')

        user_model_py = self.test_project_dir / 'models' / 'user_model.py'
        user_model_py.write_text('class UserModel(BaseModel):\n    pass\n')

        # Create files in services
        base_service_py = self.test_project_dir / 'services' / 'base_service.py'
        base_service_py.write_text('class BaseService:\n    pass\n')

        user_service_py = self.test_project_dir / 'services' / 'user_service.py'
        user_service_py.write_text('@service_decorator\nclass UserService(BaseService):\n    pass\n')

        deprecated_service_py = self.test_project_dir / 'excluded_folder_name' / 'deprecated_service.py'
        deprecated_service_py.write_text('@deprecated_service\nclass DeprecatedService(BaseService):\n    pass\n')

        # Create utility functions
        utils_py = self.test_project_dir / 'utilities' / 'utils.py'
        utils_py.write_text('@utility_decorator\ndef helper_function():\n    pass\n')

    def tearDown(self):
        # Remove temporary directory
        shutil.rmtree(self.test_dir)

    def run_collector(self, collector, paths, exclude_files):
        all_elements = set()
        for base_path_str in paths:
            base_path = Path(base_path_str)
            if not base_path.exists():
                continue
            files = [f for f in base_path.rglob("*.py") if f.is_file()]

            # Apply global exclude patterns
            def is_excluded(file_path: Path) -> bool:
                relative_path = str(file_path.relative_to(base_path))
                return any(pattern.search(relative_path) for pattern in exclude_files)

            files = [f for f in files if not is_excluded(f)]

            for f in files:
                try:
                    with open(f, "r", encoding="utf-8") as file:
                        file_content = file.read()
                    file_ast = ast.parse(file_content, filename=str(f))
                except:
                    continue

                matched = collector.match_in_file(file_ast, f)
                all_elements.update(matched)

        return all_elements

    def test_class_inherits_collector(self):
        collector_config = {'base_class': 'BaseModel'}
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = ClassInheritsCollector(collector_config)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_class_names = {element.name for element in collected_elements}
        expected_classes = {'UserModel'}
        self.assertEqual(collected_class_names, expected_classes)

    def test_file_regex_collector(self):
        collector_config = {
            'regex': r'.*controller.py$',
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = FileRegexCollector(collector_config, paths, exclude_files)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_class_names = {element.name for element in collected_elements}
        expected_classes = {'BaseController', 'UserController'}
        self.assertEqual(collected_class_names, expected_classes)

    def test_class_name_regex_collector(self):
        collector_config = {
            'class_name_regex': '^User.*',
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = ClassNameRegexCollector(collector_config, paths, exclude_files)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_class_names = {element.name for element in collected_elements}
        expected_classes = {'UserController', 'UserModel', 'UserService'}
        self.assertEqual(collected_class_names, expected_classes)

    def test_directory_collector(self):
        collector_config = {
            'directories': ['services'],
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = DirectoryCollector(collector_config, paths, exclude_files)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_class_names = {element.name for element in collected_elements}
        expected_classes = {'BaseService', 'UserService'}
        self.assertEqual(collected_class_names, expected_classes)

    def test_decorator_usage_collector(self):
        collector_config = {
            'decorator_name': 'login_required',
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = DecoratorUsageCollector(collector_config)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_names = {element.name for element in collected_elements}
        expected_names = {'UserController'}
        self.assertEqual(collected_names, expected_names)

        # Test with decorator_regex
        collector_config = {
            'decorator_regex': '^.*decorator$',
        }
        collector = DecoratorUsageCollector(collector_config)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_names = {element.name for element in collected_elements}
        expected_names = {'UserService', 'helper_function'}
        self.assertEqual(collected_names, expected_names)

    def test_class_name_regex_collector_no_matches(self):
        collector_config = {
            'class_name_regex': '^NonExistentClass.*',
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = ClassNameRegexCollector(collector_config, paths, exclude_files)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        self.assertEqual(len(collected_elements), 0)

    def test_function_name_regex_collector(self):
        extra_py = self.test_project_dir / 'utilities' / 'extra.py'
        extra_py.write_text('def helper_secondary_function():\n    pass\n')

        collector_config = {
            'function_name_regex': '^helper_.*'
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        from deply.collectors.function_name_regex_collector import FunctionNameRegexCollector
        collector = FunctionNameRegexCollector(collector_config)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_function_names = {element.name for element in collected_elements}
        expected_functions = {'helper_function', 'helper_secondary_function'}
        self.assertEqual(collected_function_names, expected_functions)

    def test_bool_collector(self):
        collector_config = {
            'type': 'bool',
            'must': [
                {'type': 'class_name_regex', 'class_name_regex': '.*Service$'}
            ],
            'must_not': [
                {'type': 'file_regex', 'regex': '.*/base_service.py'},
                {'type': 'file_regex', 'regex': '.*/excluded_folder_name/.*'},
                {'type': 'decorator_usage', 'decorator_name': 'deprecated_service'}
            ]
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []
        collector = BoolCollector(collector_config, paths, exclude_files)
        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_class_names = {element.name for element in collected_elements}
        expected_classes = {'UserService'}
        self.assertEqual(collected_class_names, expected_classes)

    def test_directory_collector_with_rules(self):
        user_controller_py = self.test_project_dir / 'controllers' / 'user_controller.py'
        user_controller_py.write_text(
            'from ..models.user_model import UserModel\n'
            'class UserController:\n'
            '    def __init__(self):\n'
            '        self.model = UserModel()\n'
        )
        config_yaml = Path(self.test_dir) / 'config_directory_collector_rules.yaml'
        config_data = {
            'deply': {
                'paths': [str(self.test_project_dir)],
                'layers': [
                    {
                        'name': 'models_layer',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['models'],
                            }
                        ]
                    },
                    {
                        'name': 'controllers_layer',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['controllers'],
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'controllers_layer': {
                        'disallow_layer_dependencies': ['models_layer']
                    }
                }
            }
        }
        with config_yaml.open('w') as f:
            yaml.dump(config_data, f)
        with self.capture_output() as (out, err):
            try:
                sys.argv = ['main.py', 'analyze', '--config', str(config_yaml)]
                main()
            except SystemExit as e:
                exit_code = e.code
        output = out.getvalue()
        self.assertEqual(exit_code, 1)
        self.assertIn("Layer 'controllers_layer' is not allowed to depend on layer 'models_layer'", output)

    def test_directory_collector_class_fields_with_fqn(self):
        (self.test_project_dir / 'services').mkdir(exist_ok=True)
        service_file = self.test_project_dir / 'services' / 'my_service.py'
        service_file.write_text(
            'class MyService:\n'
            '    path: str = "/some/path"\n'
            '    another_var = 123\n'
        )

        collector_config = {
            'directories': ['services'],
        }
        paths = [str(self.test_project_dir)]
        exclude_files = []

        from deply.collectors.directory_collector import DirectoryCollector
        collector = DirectoryCollector(collector_config, paths, exclude_files)

        collected_elements = self.run_collector(collector, paths, exclude_files)
        collected_variable_names = {
            elem.name for elem in collected_elements if elem.element_type == 'variable'
        }

        # We expect 'MyService.path' and 'MyService.another_var', not just 'path' or 'another_var'
        self.assertIn('MyService.path', collected_variable_names)
        self.assertIn('MyService.another_var', collected_variable_names)
        self.assertNotIn('path', collected_variable_names)
        self.assertNotIn('another_var', collected_variable_names)

    @staticmethod
    def capture_output():
        from contextlib import contextmanager
        from io import StringIO
        import sys

        @contextmanager
        def _capture_output():
            new_out, new_err = StringIO(), StringIO()
            old_out, old_err = sys.stdout, sys.stderr
            try:
                sys.stdout, sys.stderr = new_out, new_err
                yield sys.stdout, sys.stderr
            finally:
                sys.stdout, sys.stderr = old_out, old_err

        return _capture_output()


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from pathlib import Path

from deply.models.dependency import Dependency
from deply.models.violation_types import ViolationType


@dataclass(frozen=True)
class Violation:
    file: Path
    element_name: str
    element_type: str  # 'class', 'function', or 'variable'
    line: int
    column: int
    message: str
    violation_type: ViolationType
    dependency: Dependency = None

    def __hash__(self):
        return hash((self.file, self.line, self.column, self.message, self.violation_type))

    def __eq__(self, other):
        return (
                (self.file, self.line, self.column, self.message, self.violation_type)
                == (other.file, other.line, other.column, other.message, other.violation_type)
        )

    def to_dict(self) -> dict:
        return {
            "file": str(self.file),
            "element_name": self.element_name,
            "element_type": self.element_type,
            "line": self.line,
            "column": self.column,
            "message": self.message,
            "violation_type": self.violation_type.code,
        }



import ast
from pathlib import Path
from typing import Any, Dict, List, Set

from deply.models.code_element import CodeElement
from .base_collector import BaseCollector


class BoolCollector(BaseCollector):
    def __init__(self, config: Dict[str, Any], paths: List[str], exclude_files: List[str]):
        self.must_configs = config.get('must', [])
        self.any_of_configs = config.get('any_of', [])
        self.must_not_configs = config.get('must_not', [])

        # Pre-instantiate sub-collectors
        from .collector_factory import CollectorFactory
        self.must_collectors = [CollectorFactory.create(c, paths, exclude_files) for c in self.must_configs]
        self.any_of_collectors = [CollectorFactory.create(c, paths, exclude_files) for c in self.any_of_configs]
        self.must_not_collectors = [CollectorFactory.create(c, paths, exclude_files) for c in self.must_not_configs]

    def match_in_file(self, file_ast: ast.AST, file_path: Path) -> Set[CodeElement]:
        must_sets = []
        for c in self.must_collectors:
            must_sets.append(c.match_in_file(file_ast, file_path))
        any_of_sets = []
        for c in self.any_of_collectors:
            any_of_sets.append(c.match_in_file(file_ast, file_path))
        must_not_elements = set()
        for c in self.must_not_collectors:
            must_not_elements.update(c.match_in_file(file_ast, file_path))

        if must_sets:
            must_elements = set.intersection(*must_sets) if must_sets else set()
        else:
            must_elements = None

        if any_of_sets:
            any_of_elements = set.union(*any_of_sets) if any_of_sets else set()
        else:
            any_of_elements = None

        if must_elements is not None and any_of_elements is not None:
            combined_elements = must_elements & any_of_elements
        elif must_elements is not None:
            combined_elements = must_elements
        elif any_of_elements is not None:
            combined_elements = any_of_elements
        else:
            combined_elements = set()

        final_elements = combined_elements - must_not_elements

        return final_elements



from typing import List

from .formats.github_actions_report import GitHubActionsReport
from .formats.json_report import JsonReport
from ..models.violation import Violation
from .formats.text_report import TextReport


class ReportGenerator:
    def __init__(self, violations: List[Violation]):
        self.violations = violations

    def generate(self, format: str) -> str:
        if format == "text":
            reporter = TextReport(self.violations)
        elif format == "json":
            reporter = JsonReport(self.violations)
        elif format == 'github-actions':
            reporter = GitHubActionsReport(self.violations)
        else:
            reporter = TextReport(self.violations)

        return reporter.generate()



import json
from collections import defaultdict
from typing import List
from ...models.violation import Violation


class JsonReport:
    def __init__(self, violations: List[Violation]):
        self.violations = violations

    def generate(self) -> str:
        grouped_violations = self._group_violations_by_type()

        data = {
            "total_violations": len(self.violations),
            "by_type": {
                v_type: len(v_list) for v_type, v_list in grouped_violations.items()
            },
            "violations": [
                {
                    "file": str(violation.file),
                    "element_name": violation.element_name,
                    "element_type": violation.element_type,
                    "line": violation.line,
                    "column": violation.column,
                    "message": violation.message,
                    "violation_type": violation.violation_type.code,
                }
                for violation in self.violations
            ],
        }

        return json.dumps(data, indent=2)

    def _group_violations_by_type(self) -> dict[str, List[Violation]]:
        grouped = defaultdict(list)
        for violation in self.violations:
            grouped[violation.violation_type.code].append(violation.to_dict())
        return dict(grouped)



import re
from typing import Optional
from .base_rule import BaseRule
from ..models.violation import Violation
from ..models.code_element import CodeElement
from ..models.violation_types import ViolationType


class ClassDecoratorUsageRule(BaseRule):
    VIOLATION_TYPE = ViolationType.CLASS_DECORATOR_USAGE

    def __init__(self, layer_name: str, decorator_regex: str):
        self.layer_name = layer_name
        self.decorator_pattern = re.compile(decorator_regex)

    def check_element(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        if layer_name != self.layer_name or element.element_type != 'class':
            return None
        decorators = getattr(element, 'decorators', [])
        if any(self.decorator_pattern.match(decorator) for decorator in decorators):
            return None
        return Violation(
            file=element.file,
            element_name=element.name,
            element_type=element.element_type,
            line=element.line,
            column=element.column,
            message=(
                f"Class '{element.name}' must have a decorator matching '{self.decorator_pattern.pattern}'."
            ),
            violation_type=self.VIOLATION_TYPE
        )



from dataclasses import dataclass

from .dependency import Dependency
from ..models.code_element import CodeElement


@dataclass()
class Layer:
    name: str
    code_elements: set[CodeElement]
    dependencies: set[Dependency]



from typing import Dict, Any, List, Optional
from .base_rule import BaseRule
from .dependency_rule import DependencyRule
from .class_naming_rule import ClassNamingRule
from .function_naming_rule import FunctionNamingRule
from .class_decorator_rule import ClassDecoratorUsageRule
from .function_decorator_rule import FunctionDecoratorUsageRule
from .inheritance_rule import InheritanceRule
from .bool_rule import BoolRule


class RuleFactory:
    @staticmethod
    def create_sub_rules(rule_configs: List[Dict[str, Any]], layer_name: str) -> List[BaseRule]:
        rules = []
        for rule_config in rule_configs:
            rule = RuleFactory._create_rule_from_config(layer_name, rule_config)
            if rule is not None:
                rules.append(rule)
        return rules

    @staticmethod
    def create_rules(ruleset: Dict[str, Any]) -> List[BaseRule]:
        rules = []
        for layer_name, layer_rules in ruleset.items():
            disallowed = layer_rules.get("disallow_layer_dependencies")
            if disallowed:
                rules.append(DependencyRule(layer_name, disallowed))

            rules.extend(
                RuleFactory._collect_rules_for_key(layer_name, layer_rules, "enforce_class_naming")
            )
            rules.extend(
                RuleFactory._collect_rules_for_key(layer_name, layer_rules, "enforce_function_naming")
            )
            rules.extend(
                RuleFactory._collect_rules_for_key(layer_name, layer_rules, "enforce_class_decorator_usage")
            )
            rules.extend(
                RuleFactory._collect_rules_for_key(layer_name, layer_rules, "enforce_function_decorator_usage")
            )
            rules.extend(
                RuleFactory._collect_rules_for_key(layer_name, layer_rules, "enforce_inheritance")
            )

        return rules

    @staticmethod
    def _collect_rules_for_key(layer_name: str, layer_rules: Dict[str, Any], key: str) -> List[BaseRule]:
        if key not in layer_rules:
            return []
        configs = layer_rules[key]
        collected = []
        for rule_config in configs:
            rule = RuleFactory._create_rule_from_config(layer_name, rule_config)
            if rule is not None:
                collected.append(rule)
        return collected

    @staticmethod
    def _create_rule_from_config(layer_name: str, rule_config: Dict[str, Any]) -> Optional[BaseRule]:
        rule_type = rule_config.get("type")
        if rule_type == "class_name_regex":
            regex = rule_config.get("class_name_regex", "")
            return ClassNamingRule(layer_name, regex)
        if rule_type == "function_name_regex":
            regex = rule_config.get("function_name_regex", "")
            return FunctionNamingRule(layer_name, regex)
        if rule_type == "function_decorator_name_regex":
            regex = rule_config.get("decorator_name_regex", "")
            return FunctionDecoratorUsageRule(layer_name, regex)
        if rule_type == "class_decorator_name_regex":
            regex = rule_config.get("decorator_name_regex", "")
            return ClassDecoratorUsageRule(layer_name, regex)
        if rule_type == "class_inherits":
            base_class = rule_config.get("base_class", "")
            return InheritanceRule(layer_name, base_class)
        if rule_type == "bool":
            return BoolRule(layer_name, rule_config)
        return None



from collections import defaultdict
from typing import List

from ...models.violation import Violation


class SummaryTable:
    COL1_WIDTH = 30
    COL2_WIDTH = 5

    def __init__(self, summary_data: dict[str, int]):
        self.summary_data = summary_data

    def generate(self) -> str:
        lines = []
        lines.append(self._table_header())
        # "Report" in column 1, blank in column 2
        lines.append(self._table_row("Violations report", ""))
        lines.append(self._table_header())

        for key, value in self.summary_data.items():
            lines.append(self._table_row(key, str(value)))

        lines.append(self._table_header())
        return "\n".join(lines)

    def _table_header(self) -> str:
        return (
                " " + "-" * self.COL1_WIDTH + " " +
                "-" * self.COL2_WIDTH + " "
        )

    def _table_row(self, col1: str, col2: str) -> str:
        col1 = col1[:self.COL1_WIDTH]
        col2 = col2[:self.COL2_WIDTH]
        return (
                " "
                + col1.ljust(self.COL1_WIDTH)
                + " "
                + col2.ljust(self.COL2_WIDTH)
                + " "
        )


class TextReport:
    def __init__(self, violations: List[Violation]):
        self.violations = violations

    def generate(self) -> str:
        grouped_violations = self._group_violations_by_type()
        lines = []

        # 1) Print violations grouped by their display name
        for violation_type, type_violations in grouped_violations.items():
            lines.append(violation_type)
            sorted_violations = sorted(type_violations, key=lambda v: (v.file, v.line, v.column))
            for violation in sorted_violations:
                lines.append(f"{violation.file}:{violation.line}:{violation.column} - {violation.message}")
            lines.append("")

        # 2) Build a dict of summary data
        summary_data = {}
        for violation_type, type_violations in grouped_violations.items():
            summary_data[violation_type] = len(type_violations)

        # Add a total
        total_count = len(self.violations)
        summary_data[""] = ""
        summary_data["Total Violations"] = total_count

        # 3) Generate the summary table
        summary_table = SummaryTable(summary_data)
        lines.append(summary_table.generate())

        return "\n".join(lines)

    def _group_violations_by_type(self) -> dict[str, List[Violation]]:
        grouped = defaultdict(list)
        for violation in self.violations:
            grouped[violation.violation_type.display_name].append(violation)
        return dict(grouped)



from enum import Enum


class ViolationType(Enum):
    DISALLOWED_DEPENDENCY = ("disallowed_dependency", "Disallowed Dependency")
    FUNCTION_DECORATOR_USAGE = ("function_decorator_usage", "Function Decorator Usage")
    CLASS_DECORATOR_USAGE = ("class_decorator_usage", "Class Decorator Usage")
    CLASS_NAMING = ("class_naming", "Class Naming")
    FUNCTION_NAMING = ("function_naming", "Function Naming")
    INHERITANCE = ("inheritance", "Inheritance")
    BOOL_RULE = ("bool_rule", "Bool Rule")

    def __init__(self, code: str, display_name: str):
        self._code = code
        self._display_name = display_name

    def to_dict(self) -> dict:
        return {
            "code": self.code,
            "display_name": self.display_name
        }

    @property
    def code(self) -> str:
        return self._code

    @property
    def display_name(self) -> str:
        return self._display_name



import re
from typing import Optional
from .base_rule import BaseRule
from ..models.violation import Violation
from ..models.code_element import CodeElement
from ..models.violation_types import ViolationType


class ClassNamingRule(BaseRule):
    VIOLATION_TYPE = ViolationType.CLASS_NAMING

    def __init__(self, layer_name: str, regex: str):
        self.layer_name = layer_name
        self.class_name_pattern = re.compile(regex)

    def check_element(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        if layer_name != self.layer_name or element.element_type != 'class':
            return None
        class_name = element.name.split('.')[-1]
        if self.class_name_pattern.match(class_name):
            return None
        return Violation(
            file=element.file,
            element_name=element.name,
            element_type=element.element_type,
            line=element.line,
            column=element.column,
            message=(
                f"Class '{element.name}' does not match naming pattern '{self.class_name_pattern.pattern}'."
            ),
            violation_type=self.VIOLATION_TYPE
        )



from collections import defaultdict
from typing import List

from ...models.violation import Violation


class GitHubActionsReport:
    def __init__(self, violations: List[Violation]):
        self.violations = violations

    def generate(self) -> str:
        grouped_violations = self._group_violations_by_type()
        lines = []
        for violation_type, type_violations in grouped_violations.items():
            sorted_violations = sorted(type_violations, key=lambda v: (v.file, v.line, v.column))
            for violation in sorted_violations:
                # Using warning format. Could be "::error" or "::warning"
                # Format: ::warning file=<file>,line=<line>,col=<column>::<message>
                lines.append(
                    f"::warning file={violation.file},line={violation.line},col={violation.column}::{violation.message}"
                )

        # Add summary at the end as a comment
        total_count = len(self.violations)
        for violation_type, type_violations in grouped_violations.items():
            lines.append(f"# {violation_type}: {len(type_violations)}")
        lines.append(f"# Total Violations: {total_count}")

        return "\n".join(lines)

    def _group_violations_by_type(self) -> dict[str, List[Violation]]:
        grouped = defaultdict(list)
        for violation in self.violations:
            grouped[violation.violation_type.display_name].append(violation)
        return dict(grouped)



from typing import Optional

from deply.models.violation_types import ViolationType
from deply.rules.base_rule import BaseRule
from deply.models.code_element import CodeElement
from deply.models.violation import Violation


class InheritanceRule(BaseRule):
    VIOLATION_TYPE = ViolationType.INHERITANCE

    def __init__(self, layer_name: str, base_class: str):
        self.layer_name = layer_name
        self.base_class = base_class

    def check_element(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        if layer_name != self.layer_name or element.element_type != 'class':
            return None
        if any(
                inherited_class == self.base_class or inherited_class.endswith(f".{self.base_class}")
                for inherited_class in element.inherits
        ):
            return None
        return Violation(
            file=element.file,
            element_name=element.name,
            element_type=element.element_type,
            line=element.line,
            column=element.column,
            message=(
                f"Class '{element.name}' must inherit from '{self.base_class}'."
            ),
            violation_type=self.VIOLATION_TYPE
        )



from typing import Dict, Any, List
from .base_collector import BaseCollector
from .class_inherits_collector import ClassInheritsCollector
from .class_name_regex_collector import ClassNameRegexCollector
from .directory_collector import DirectoryCollector
from .file_regex_collector import FileRegexCollector
from .decorator_usage_collector import DecoratorUsageCollector
from .bool_collector import BoolCollector
from .function_name_regex_collector import FunctionNameRegexCollector

class CollectorFactory:
    @staticmethod
    def create(config: Dict[str, Any], paths: List[str], exclude_files: List[str]) -> BaseCollector:
        collector_type = config.get("type")
        if collector_type == "file_regex":
            return FileRegexCollector(config, paths, exclude_files)
        elif collector_type == "class_inherits":
            return ClassInheritsCollector(config)
        elif collector_type == "class_name_regex":
            return ClassNameRegexCollector(config, paths, exclude_files)
        elif collector_type == "function_name_regex":
            return FunctionNameRegexCollector(config)
        elif collector_type == "directory":
            return DirectoryCollector(config, paths, exclude_files)
        elif collector_type == "decorator_usage":
            return DecoratorUsageCollector(config)
        elif collector_type == "bool":
            return BoolCollector(config, paths, exclude_files)
        else:
            raise ValueError(f"Unknown collector type: {collector_type}")


from typing import Dict, Any, Optional
from deply.models.code_element import CodeElement
from deply.models.violation import Violation
from deply.models.violation_types import ViolationType
from deply.rules.base_rule import BaseRule


class BoolRule(BaseRule):
    VIOLATION_TYPE = ViolationType.BOOL_RULE

    def __init__(self, layer_name: str, config: Dict[str, Any]):
        self.layer_name = layer_name
        self.must_configs = config.get('must', [])
        self.any_of_configs = config.get('any_of', [])
        self.must_not_configs = config.get('must_not', [])

        from deply.rules.rule_factory import RuleFactory
        self.must_rules = RuleFactory.create_sub_rules(self.must_configs, layer_name)
        self.any_of_rules = RuleFactory.create_sub_rules(self.any_of_configs, layer_name)
        self.must_not_rules = RuleFactory.create_sub_rules(self.must_not_configs, layer_name)

    def check_element(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        if layer_name != self.layer_name:
            return None

        must_violation = self._check_must_rules(layer_name, element)
        if must_violation:
            return must_violation

        any_of_violation = self._check_any_of_rules(layer_name, element)
        if any_of_violation:
            return any_of_violation

        must_not_violation = self._check_must_not_rules(layer_name, element)
        if must_not_violation:
            return must_not_violation

        return None

    def _check_must_rules(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        for rule in self.must_rules:
            violation = rule.check_element(layer_name, element)
            if violation:
                return violation
        return None

    def _check_any_of_rules(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        if not self.any_of_rules:
            return None

        passed = False
        last_violation = None
        for rule in self.any_of_rules:
            violation = rule.check_element(layer_name, element)
            if violation is None:
                passed = True
                break
            last_violation = violation

        if not passed and last_violation:
            return last_violation

        return None

    def _check_must_not_rules(self, layer_name: str, element: CodeElement) -> Optional[Violation]:
        for rule in self.must_not_rules:
            violation = rule.check_element(layer_name, element)
            if violation is None:
                return Violation(
                    file=element.file,
                    element_name=element.name,
                    element_type=element.element_type,
                    line=element.line,
                    column=element.column,
                    message=(
                        f"BoolRule failed: must_not rule {rule.__class__.__name__} "
                        f"did not produce a violation."
                    ),
                    violation_type=self.VIOLATION_TYPE
                )
        return None



# deply/models/code_element.py
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Tuple, FrozenSet


@dataclass(frozen=True)
class CodeElement:
    file: Path
    name: str  # Should include fully qualified name if possible
    element_type: str  # 'class', 'function', or 'variable'
    line: int
    column: int

    decorators: Tuple[str, ...] = field(default_factory=tuple)
    inherits: Tuple[str, ...] = field(default_factory=tuple)
    return_annotation: Optional[str] = None
    type_annotations: FrozenSet[Tuple[str, str]] = field(default_factory=frozenset)



import ast
import re
from pathlib import Path
from typing import List, Set, Dict

from deply.collectors import BaseCollector
from deply.models.code_element import CodeElement
from deply.utils.ast_utils import get_import_aliases, get_base_name, get_decorator_name, get_annotation_name


class ClassNameRegexCollector(BaseCollector):
    def __init__(self, config: dict, paths: List[str], exclude_files: List[str]):
        self.regex_pattern = config.get("class_name_regex", "")
        self.exclude_files_regex_pattern = config.get("exclude_files_regex", "")
        self.regex = re.compile(self.regex_pattern)
        self.exclude_regex = re.compile(self.exclude_files_regex_pattern) if self.exclude_files_regex_pattern else None

        self.paths = [Path(p) for p in paths]
        self.exclude_files = [re.compile(pattern) for pattern in exclude_files]

    def match_in_file(self, file_ast: ast.AST, file_path: Path) -> Set[CodeElement]:
        if self.exclude_regex and self.exclude_regex.search(str(file_path)):
            return set()

        import_aliases = get_import_aliases(file_ast)
        classes = set()

        for node in ast.walk(file_ast):
            if isinstance(node, ast.ClassDef):
                if self.regex.match(node.name):
                    # Gather inherits
                    inherits_list = []
                    for base in node.bases:
                        base_name = get_base_name(base, import_aliases)
                        inherits_list.append(base_name)

                    # Gather decorators
                    decorators_list = []
                    for d in node.decorator_list:
                        dec_name = get_decorator_name(d)
                        if dec_name is not None:
                            decorators_list.append(dec_name)

                    # Gather type annotations (class-level attributes)
                    type_annotations: Dict[str, str] = {}
                    for stmt in node.body:
                        if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                            ann_name = get_annotation_name(stmt.annotation, import_aliases)
                            if ann_name is not None:
                                type_annotations[stmt.target.id] = ann_name

                    full_name = self._get_full_name(node)
                    code_element = CodeElement(
                        file=file_path,
                        name=full_name,
                        element_type='class',
                        line=node.lineno,
                        column=node.col_offset,
                        inherits=tuple(inherits_list),
                        decorators=tuple(decorators_list),
                        return_annotation=None,  # Classes do not have return annotations
                        type_annotations=frozenset(type_annotations.items())
                    )
                    classes.add(code_element)
        return classes

    def _get_full_name(self, node):
        names = []
        current = node
        while isinstance(current, (ast.ClassDef, ast.FunctionDef)):
            names.append(current.name)
            current = getattr(current, 'parent', None)
        return '.'.join(reversed(names))



from dataclasses import dataclass

from deply.models.code_element import CodeElement


@dataclass(frozen=True)
class Dependency:
    code_element: CodeElement
    depends_on_code_element: CodeElement
    dependency_type: str
    line: int
    column: int



class MermaidDiagramBuilder:
    def __init__(self):
        self._edges_with_violation: dict[tuple[str, str], bool] = {}

    def add_edge(self, source_layer: str, target_layer: str, has_violation: bool) -> None:
        """
        Adds or updates an edge (source_layer -> target_layer) with a flag indicating
        if a violation occurred. If we’ve already recorded a violation for this edge,
        we keep it as True.
        """
        current_violation = self._edges_with_violation.get((source_layer, target_layer), False)
        self._edges_with_violation[(source_layer, target_layer)] = current_violation or has_violation

    def build_diagram(self) -> str:
        """
        Builds and returns a Mermaid flowchart (as a Markdown code block) that
        shows all layers. Arrows are colored red (via linkStyle) if has_violation=True.
        """
        # Gather unique layers
        layer_set = set()
        for (source, target), _ in self._edges_with_violation.items():
            layer_set.add(source)
            layer_set.add(target)

        # Sort edges by source_layer, then target_layer for consistent output
        sorted_edges = sorted(self._edges_with_violation.items(), key=lambda kv: (kv[0][0], kv[0][1]))
        lines = ["```mermaid", "flowchart LR"]

        # Ensure every layer is at least mentioned so it appears in the diagram
        for layer in sorted(layer_set):
            lines.append(f"    {layer}")

        # We'll collect linkStyle lines separately
        link_style_lines = []
        edge_index = 0

        # Build edges
        for (source, target), violated in sorted_edges:
            lines.append(f"    {source} --> {target}")
            if violated:
                # Mark arrow red
                link_style_lines.append(f"    linkStyle {edge_index} stroke:red,stroke-width:2px;")
            edge_index += 1

        lines.extend(link_style_lines)
        lines.append("```")
        return "\n".join(lines)

    @property
    def edges_with_violation(self) -> dict[tuple[str, str], bool]:
        return self._edges_with_violation



from pathlib import Path
from typing import Any, Dict
import yaml


class ConfigParser:
    def __init__(self, config_path: Path):
        self.config_path = config_path

    def parse(self) -> Dict[str, Any]:
        with self.config_path.open("r") as f:
            config = yaml.safe_load(f)

        config = config.get('deply', config)
        config.setdefault('paths', [])
        config.setdefault('exclude_files', [])
        config.setdefault('layers', [])
        config.setdefault('ruleset', {})

        if not config['paths']:
            config['paths'] = [str(self.config_path.parent)]

        return config



import argparse
import ast
import logging
import re
import sys
from pathlib import Path

from deply import __version__
from deply.diagrams.marmaid_diagram_builder import MermaidDiagramBuilder
from deply.reports.report_generator import ReportGenerator
from deply.rules.rule_factory import RuleFactory
from .code_analyzer import CodeAnalyzer
from .collectors.collector_factory import CollectorFactory
from .config_parser import ConfigParser
from .models.code_element import CodeElement
from .models.layer import Layer
from .models.violation import Violation


def main():
    parser = argparse.ArgumentParser(prog="deply", description='Deply - An architecture analysis tool')
    parser.add_argument('-V', '--version', action='store_true', help='Show the version number and exit')
    parser.add_argument('-v', '--verbose', action='count', default=1, help='Increase output verbosity')

    subparsers = parser.add_subparsers(dest='command', help='Sub-commands')
    parser_analyze = subparsers.add_parser('analyze', help='Analyze the project')
    parser_analyze.add_argument('--config', type=str, default="deply.yaml", help="Path to the configuration YAML file")
    parser_analyze.add_argument(
        '--report-format',
        type=str,
        choices=["text", "json", "github-actions"],
        default="text",
        help="Format of the output report"
    )
    parser_analyze.add_argument('--output', type=str, help="Output file for the report")
    parser_analyze.add_argument('--mermaid', action='store_true',
                                help="Generate a Mermaid diagram for layer dependencies (red = violation)")

    args = parser.parse_args()

    if args.version:
        version = __version__
        print(f"deply {version}")
        sys.exit(0)

    # Set up logging
    log_level = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG

    logging.basicConfig(
        level=log_level,
        format='[%(asctime)s] [%(levelname)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    if args.command is None:
        args = parser.parse_args(['analyze'] + sys.argv[1:])

    logging.info("Starting Deply analysis...")

    config_path = Path(args.config)
    logging.info(f"Using configuration file: {config_path}")
    config = ConfigParser(config_path).parse()

    paths = [Path(p) for p in config["paths"]]
    exclude_files = [re.compile(pattern) for pattern in config["exclude_files"]]
    layers_config = config["layers"]
    ruleset = config["ruleset"]

    logging.info("Mapping layer collectors...")
    layer_collectors = []
    for layer_config in layers_config:
        layer_name = layer_config["name"]
        collector_configs = layer_config.get("collectors", [])
        for collector_config in collector_configs:
            collector = CollectorFactory.create(
                config=collector_config,
                paths=[str(p) for p in paths],
                exclude_files=[p.pattern for p in exclude_files]
            )
            layer_collectors.append((layer_name, collector))

    logging.info("Collecting all files...")
    all_files = []
    for base_path in paths:
        if not base_path.exists():
            continue
        all_python_files = [f for f in base_path.rglob("*.py") if f.is_file()]

        def is_excluded(file_path: Path) -> bool:
            try:
                relative_path = str(file_path.relative_to(base_path))
            except ValueError:
                return True
            return any(pattern.search(relative_path) for pattern in exclude_files)

        filtered_files = [f for f in all_python_files if not is_excluded(f)]
        all_files.extend(filtered_files)

    layers: dict[str, Layer] = {}
    for layer_config in layers_config:
        layer_name = layer_config["name"]
        layers[layer_name] = Layer(name=layer_name, code_elements=set(), dependencies=set())

    code_element_to_layer: dict[CodeElement, str] = {}
    logging.info("Collecting code elements for each layer...")
    for file_path in all_files:
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                file_content = f.read()
            file_ast = ast.parse(file_content, filename=str(file_path))
        except Exception as ex:
            logging.debug(f"Skipping file {file_path} due to parse error: {ex}")
            continue

        for layer_name, collector in layer_collectors:
            matched_elements = collector.match_in_file(file_ast, file_path)
            for element in matched_elements:
                layers[layer_name].code_elements.add(element)
                code_element_to_layer[element] = layer_name

    for layer_name, layer in layers.items():
        logging.info(f"Layer '{layer_name}' collected {len(layer.code_elements)} code elements.")

    logging.info("Preparing rules...")
    rules = RuleFactory.create_rules(ruleset)

    violations: set[Violation] = set()
    metrics = {'total_dependencies': 0}

    mermaid_builder = MermaidDiagramBuilder()

    def dependency_handler(dependency):
        source = dependency.code_element
        target = dependency.depends_on_code_element
        source_layer = code_element_to_layer.get(source)
        target_layer = code_element_to_layer.get(target)
        metrics['total_dependencies'] += 1

        if not source_layer or not target_layer:
            return
        if source_layer == target_layer:
            # Intra-layer dependency, not relevant for cross-layer analysis
            return

        # Check if there's a violation for this dependency
        has_violation = False
        for rule in rules:
            violation = rule.check(source_layer, target_layer, dependency)
            if violation:
                violations.add(violation)
                has_violation = True

        # Pass edge info to mermaid builder
        mermaid_builder.add_edge(source_layer, target_layer, has_violation)

    logging.info("Analyzing code and checking dependencies ...")
    analyzer = CodeAnalyzer(
        code_elements=set(code_element_to_layer.keys()),
        dependency_handler=dependency_handler
    )
    analyzer.analyze()

    logging.info(f"Analysis complete. Found {metrics['total_dependencies']} dependencies(s).")

    logging.info("Running element-based checks ...")
    for layer_name, layer in layers.items():
        for element in layer.code_elements:
            for rule in rules:
                v = rule.check_element(layer_name, element)
                if v:
                    violations.add(v)

    logging.info("Generating report...")
    report = ReportGenerator(list(violations)).generate(args.report_format)

    if args.output:
        output_path = Path(args.output)
        output_path.write_text(report)
        logging.info(f"Report written to {output_path}")
    else:
        print("\n")
        print(report)

    if args.mermaid:
        mermaid_diagram = mermaid_builder.build_diagram()
        print("\n[Mermaid Diagram of Layer Dependencies]\n")
        print(mermaid_diagram)

    if violations:
        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    main()



from deply import __version__
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="deply",
    version=__version__,
    author="Archil Abuladze",
    author_email="armiworker@gmail.com",
    description="A tool to enforce architectural patterns in python projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vashkatsi/deply",
    project_urls={
        "Bug Tracker": "https://github.com/vashkatsi/deply/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=[
        "PyYAML>=5.1",
    ],
    entry_points={
        'console_scripts': [
            'deply=deply.main:main',
        ],
    },
    include_package_data=True,
)



import os
import shutil
import sys
import tempfile
import unittest
from contextlib import contextmanager
from io import StringIO
from pathlib import Path

import yaml

from deply.main import main


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestFalseDependencyPath(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.test_project_dir = Path(self.test_dir) / 'test_project'
        self.test_project_dir.mkdir()

        # Original files for the existing test_path_not_flagged_as_false_dependency
        service_py = self.test_project_dir / 'service.py'
        service_py.write_text(
            'from dataclasses import dataclass, field\n'
            'class LinkParams:\n'
            '    path: str = field(default="")\n'
            '    query: dict[str, str] = field(default_factory=dict)\n'
        )

        urls_py = self.test_project_dir / 'urls.py'
        urls_py.write_text(
            'from django.urls import path\n'
            '\n'
            'def create_path():\n'
            '    return path("some/url", None)\n'
        )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def run_deply(self, config_data) -> tuple[int, str]:
        config_path = Path(self.test_dir) / 'deply.yaml'
        with config_path.open('w') as f:
            yaml.dump(config_data, f)

        old_cwd = os.getcwd()
        os.chdir(self.test_dir)
        try:
            with captured_output() as (out, err):
                exit_code = 0
                try:
                    sys.argv = ['main.py', 'analyze', '--config', str(config_path)]
                    main()
                except SystemExit as e:
                    exit_code = e.code
            output = out.getvalue()
        finally:
            os.chdir(old_cwd)

        return exit_code, output

    def test_path_not_flagged_as_false_dependency(self):
        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'service',
                        'collectors': [
                            {
                                'type': 'file_regex',
                                'regex': '.*/service.py'
                            }
                        ]
                    },
                    {
                        'name': 'urls',
                        'collectors': [
                            {
                                'type': 'file_regex',
                                'regex': '.*/urls.py'
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'urls': {
                        'disallow_layer_dependencies': ['service']
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        self.assertEqual(exit_code, 0, f"Expected no violations. Output:\n{output}")

    def test_directory_collector_no_false_dependency(self):
        """
        This new test uses DirectoryCollector on 'services' and 'urls' directories
        to ensure referencing django.urls.path doesn't conflict with a local 'path'
        field in the service layer.
        """
        services_dir = self.test_project_dir / 'services'
        services_dir.mkdir(exist_ok=True)
        (services_dir / '__init__.py').write_text('')  # optional if needed

        my_service_file = services_dir / 'my_service.py'
        my_service_file.write_text(
            'class MyService:\n'
            '    path: str = "/test-path"\n'
            '    def do_something(self):\n'
            '        return self.path\n'
        )

        urls_dir = self.test_project_dir / 'urls_collector'
        urls_dir.mkdir(exist_ok=True)
        (urls_dir / '__init__.py').write_text('')  # optional if needed

        my_urls_file = urls_dir / 'my_urls.py'
        my_urls_file.write_text(
            'from django.urls import path\n'
            'def create_path():\n'
            '    return path("some/url", None)\n'
        )

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'services_layer',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['services']
                            }
                        ]
                    },
                    {
                        'name': 'urls_layer',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['urls_collector']
                            }
                        ]
                    }
                ],
                # Disallow the 'urls_layer' from depending on 'services_layer'
                'ruleset': {
                    'urls_layer': {
                        'disallow_layer_dependencies': ['services_layer']
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        # We expect no violations, because referencing django.urls.path
        # must not be confused with MyService.path
        self.assertEqual(exit_code, 0, f"Expected no violations. Output:\n{output}")

    def test_directory_collector_with_dependent_variable_and_directory_collector(self):
        """
        This new test uses DirectoryCollector on 'services' and 'urls' directories
        to ensure referencing django.urls.path doesn't conflict with a local 'path'
        field in the service layer.
        """
        services_dir = self.test_project_dir / 'services'
        services_dir.mkdir(exist_ok=True)
        (services_dir / '__init__.py').write_text('')  # optional if needed

        my_service_file = services_dir / 'my_service.py'
        my_service_file.write_text(
            'class MyService:\n'
            '    path: str = "/test-path"\n'
            '    def do_something(self):\n'
            '        return self.path\n'
            'test_var = 1\n'
        )

        urls_dir = self.test_project_dir / 'urls_collector'
        urls_dir.mkdir(exist_ok=True)
        (urls_dir / '__init__.py').write_text('')  # optional if needed

        my_urls_file = urls_dir / 'my_urls.py'
        my_urls_file.write_text(
            'from django.urls import path\n'
            'from services import test_var\n'
            'def create_path():\n'
            '    return path("some/url", None)\n'
        )

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'services_layer',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['services']
                            }
                        ]
                    },
                    {
                        'name': 'urls_layer',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['urls_collector']
                            }
                        ]
                    }
                ],
                # Disallow the 'urls_layer' from depending on 'services_layer'
                'ruleset': {
                    'urls_layer': {
                        'disallow_layer_dependencies': ['services_layer']
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        # We expect no violations, because referencing django.urls.path
        # must not be confused with MyService.path
        self.assertEqual(exit_code, 1, f"Expected no violations. Output:\n{output}")
        self.assertTrue(
            "test_project/urls_collector/my_urls.py:2:0 - Layer 'urls_layer' is not allowed to depend on layer 'services_layer'. Dependency type: import_from." in output,
            f"Expected no violations. Output:\n{output}"
        )


if __name__ == '__main__':
    unittest.main()



from .base_collector import BaseCollector
from .file_regex_collector import FileRegexCollector
from .class_inherits_collector import ClassInheritsCollector
from .class_name_regex_collector import ClassNameRegexCollector
from .directory_collector import DirectoryCollector
from .decorator_usage_collector import DecoratorUsageCollector



import unittest
import tempfile
import shutil
import os
from pathlib import Path
import sys
import yaml
from contextlib import contextmanager
from io import StringIO

from deply.main import main


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestRules(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()
        self.test_project_dir = Path(self.test_dir) / 'test_project'
        self.test_project_dir.mkdir()

    def tearDown(self):
        # Remove temporary directory
        shutil.rmtree(self.test_dir)

    def run_deply(self, config_data):
        # Write config
        config_yaml = Path(self.test_dir) / 'deply.yaml'
        with config_yaml.open('w') as f:
            yaml.dump(config_data, f)

        old_cwd = os.getcwd()
        os.chdir(self.test_dir)
        try:
            with captured_output() as (out, err):
                exit_code = 0
                try:
                    sys.argv = ['main.py', 'analyze', '--config', str(config_yaml)]
                    main()
                except SystemExit as e:
                    exit_code = e.code
            output = out.getvalue()
        finally:
            os.chdir(old_cwd)

        return exit_code, output

    def test_bool_rule_any_of_pass(self):
        # Create necessary files
        models_dir = self.test_project_dir / 'models'
        models_dir.mkdir()
        (models_dir / 'base_model.py').write_text('class BaseModel:\n    pass\n')
        (models_dir / 'my_model.py').write_text(
            'from .base_model import BaseModel\nclass MyModel(BaseModel):\n    pass\n')

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'models',
                        # Use file_regex to only pick up my_model.py, excluding base_model.py
                        'collectors': [
                            {
                                'type': 'file_regex',
                                'regex': '.*/my_model.py$'
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'models': {
                        'enforce_inheritance': [
                            {
                                'type': 'bool',
                                'any_of': [
                                    {
                                        'type': 'class_inherits',
                                        'base_class': 'BaseModel'
                                    },
                                    {
                                        'type': 'class_inherits',
                                        'base_class': 'AnotherModel'
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        # Now only MyModel is considered. It inherits from BaseModel, so any_of should pass.
        self.assertEqual(exit_code, 0, f"Expected no violations, but got exit code {exit_code}\nOutput:\n{output}")

    def test_bool_rule_any_of_fail(self):
        # This test checks a bool rule with any_of that should fail if none of the conditions are met.
        # We define classes that do not inherit from any required base class, expecting a violation.
        models_dir = self.test_project_dir / 'models'
        models_dir.mkdir()
        (models_dir / 'base_model.py').write_text('class BaseModel:\n    pass\n')
        (models_dir / 'my_model.py').write_text('class MyModel:\n    pass\n')

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'models',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['models']
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'models': {
                        'enforce_inheritance': [
                            {
                                'type': 'bool',
                                'any_of': [
                                    {
                                        'type': 'class_inherits',
                                        'base_class': 'BaseModel'
                                    },
                                    {
                                        'type': 'class_inherits',
                                        'base_class': 'AnotherModel'
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        # MyModel does not inherit from BaseModel or AnotherModel, so any_of fails
        self.assertEqual(exit_code, 1, f"Expected violations, but got exit code {exit_code}\nOutput:\n{output}")
        self.assertIn("Class 'MyModel' must inherit from", output)

    def test_bool_rule_must_pass(self):
        # must means all conditions must pass (no violations on sub-rules).
        # Here we will check if class_name_regex passes.
        models_dir = self.test_project_dir / 'models'
        models_dir.mkdir()
        (models_dir / 'model_a.py').write_text('class GoodModel:\n    pass\n')

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'models',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['models']
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'models': {
                        'enforce_class_naming': [
                            {
                                'type': 'bool',
                                'must': [
                                    {
                                        'type': 'class_name_regex',
                                        'class_name_regex': '^Good.*'
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        # GoodModel matches ^Good.* regex, so must passes with no violations
        self.assertEqual(exit_code, 0, f"Expected no violations, but got exit code {exit_code}\nOutput:\n{output}")

    def test_bool_rule_must_fail(self):
        # must means all conditions must pass. If one fails, we should get a violation.
        # Let's break the rule by having a class that doesn't match the regex.
        models_dir = self.test_project_dir / 'models'
        models_dir.mkdir()
        (models_dir / 'model_a.py').write_text('class BadModel:\n    pass\n')

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'models',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['models']
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'models': {
                        'enforce_class_naming': [
                            {
                                'type': 'bool',
                                'must': [
                                    {
                                        'type': 'class_name_regex',
                                        'class_name_regex': '^Good.*'
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }

        exit_code, output = self.run_deply(config_data)
        self.assertEqual(exit_code, 1, f"Expected violations, but got exit code {exit_code}\nOutput:\n{output}")
        self.assertIn("Class 'BadModel' does not match naming pattern", output)

    def test_bool_rule_must_not_pass(self):
        # must_not means all must fail (all must produce violations).
        # We'll introduce a scenario where a function should not match a regex.
        # If it matches (no violation), then must_not fails.
        models_dir = self.test_project_dir / 'models'
        models_dir.mkdir()
        (models_dir / 'model_a.py').write_text('def helper_function():\n    pass\n')

        # We want must_not = function_name_regex '^helper_.*' meaning this rule must produce violation.
        # But since 'helper_function' matches, the function_name_regex rule by itself wouldn't produce violation.
        # Actually, the function_name_regex rule produces no violation if it matches. To force a violation,
        # we need to think carefully: If a sub-rule does not produce a violation, that means it "passed".
        # must_not wants all to fail. Let's pick a rule that always fails if matched:
        # Actually, we might pick a class_inherits rule that fails since no class inherits from something.
        # Let's use a class_inherits with a base_class that doesn't exist:
        # This will produce violation on a function element because it doesn't apply. 
        # Actually, class_inherits probably returns no violation if doesn't match?
        # Let's do a simpler approach: We'll pick a sub-rule that always fails. 
        # Suppose function_name_regex means that if a function doesn't match the regex, it's a violation:
        # Actually function_name_regex rule expects the function name to match to avoid violation. If it doesn't match, violation.
        # We need a scenario that ensures a violation. Let's pick a naming rule that does not match:
        # We'll have must_not: a rule that always fails (like class_name_regex that doesn't match on a function?). 
        # Wait, we must remember the BoolRule logic: For must_not rules, we actually want violations from sub-rules. 
        # If sub-rule is function_name_regex '^helper_.*' and we do have helper_function, that sub-rule doesn't produce violation, it passes. We need it to fail.
        # Let's rename the function to break the regex:
        (models_dir / 'model_a.py').write_text('def another_func():\n    pass\n')

        # Now function_name_regex '^helper_.*' will produce a violation because 'another_func' does not match 'helper_.*'
        # So must_not rule expects a violation. We get violation from the sub-rule, meaning the sub-rule fails as intended, no violation from BoolRule itself.
        # Actually, if must_not expects all to fail (all to produce violation), and our sub-rule function_name_regex 'helper_.*' doesn't match 'another_func',
        # that means sub-rule produce a violation = good for must_not (no BoolRule violation).

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'models',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['models']
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'models': {
                        'enforce_function_naming': [
                            {
                                'type': 'bool',
                                'must_not': [
                                    {
                                        'type': 'function_name_regex',
                                        'function_name_regex': '^helper_.*'
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }

        # 'another_func' does not match '^helper_.*', so function_name_regex rule will produce a violation.
        # must_not wants violations, so this is good. The BoolRule should return no violation overall.
        exit_code, output = self.run_deply(config_data)
        self.assertEqual(exit_code, 0, f"Expected no violations, but got exit code {exit_code}\nOutput:\n{output}")

    def test_bool_rule_must_not_fail(self):
        # must_not means all must fail (produce violation).
        # If sub-rule does not produce violation, must_not fails.
        # If we have a function_name_regex '^helper_.*' and 'helper_function' matches it,
        # then function_name_regex won't produce violation, so must_not fails.

        models_dir = self.test_project_dir / 'models'
        models_dir.mkdir()
        (models_dir / 'model_a.py').write_text('def helper_function():\n    pass\n')

        config_data = {
            'deply': {
                'paths': ['./test_project'],
                'layers': [
                    {
                        'name': 'models',
                        'collectors': [
                            {
                                'type': 'directory',
                                'directories': ['models']
                            }
                        ]
                    }
                ],
                'ruleset': {
                    'models': {
                        'enforce_function_naming': [
                            {
                                'type': 'bool',
                                'must_not': [
                                    {
                                        'type': 'function_name_regex',
                                        'function_name_regex': '^helper_.*'
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }

        # 'helper_function' matches '^helper_.*', so function_name_regex produces no violation (it passes).
        # must_not wants a violation, not a pass. So BoolRule should produce a violation.
        exit_code, output = self.run_deply(config_data)
        self.assertEqual(exit_code, 1, f"Expected violations, but got exit code {exit_code}\nOutput:\n{output}")
        self.assertIn("BoolRule failed: must_not rule", output)


if __name__ == '__main__':
    unittest.main()
